#include "TMy3d.h"
/* TMy3d.cpp
   Copyright (C) 1998 ABiSoft, Inc. (bedev@abisoft.com)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

void add_eye(TShape *a_shape,
				 float center_x,
				 float center_y,
				 float center_z,
				 float size);
rgb_color make_color(uint8 r, uint8 g, uint8 b);
BPoint TMy3d::mouse_point = B_ORIGIN;

const long CIR_GRAIN = 10;
const long SPH_GRAIN = 6;

/*------------------------------------------------------------*/

TMy3d::TMy3d(BRect r, long flags, long size_x, long size_y) :
			T3d(r, flags, size_x, size_y)
{
	InitFunc();
}

/*------------------------------------------------------------*/

TMy3d::~TMy3d()
{

	dying = true;

	kill_thread(mouse_watch);
}

/*------------------------------------------------------------*/

void TMy3d::InitFunc()
{
	
	SetViewColor(0,0,0,255);
	
	alpha = beta = gamma = 0;
	_getmouse = false;
	dying = false;
	
	idle_count = 0;
	
	mouse_watch = spawn_thread( _mouse_watcher, "mouse watcher",
			 B_NORMAL_PRIORITY, this );
	resume_thread( mouse_watch );
	
	
	TShape *a_shape = new TShape(this);	
	add_eye(a_shape, 0, 0, 0, 100);

	SetShape(a_shape);
	rect = Bounds();
}

/*------------------------------------------------------------*/
//The replicant functions TMy3d, Archive, and Instantiate were
//placed here when the BeEyes were able to be replicated separately.
//They are still here for convenience of replicating one eye.

TMy3d::TMy3d(BMessage *mssg) : T3d(mssg)
{
	InitFunc();
}

/*------------------------------------------------------------*/

status_t TMy3d::Archive(BMessage *data, bool) const
{
	T3d::Archive(data);
	data->AddString("class", "TMy3d");
 	data->AddString("add_on", "application/x-vnd.abisoft-BeEyes");
 	return B_OK;
}
/*------------------------------------------------------------*/

TMy3d *
TMy3d::Instantiate(BMessage *data)
{
	if (!validate_instantiation(data, "TMy3d"))	
		return NULL;
	return new TMy3d(data);
}  

/*------------------------------------------------------------*/
void TMy3d::MouseDown(BPoint where)
{
	uint32 buttons;
	GetMouse(&where, & buttons, true);
	while(buttons) {
		GetMouse(&where, & buttons, true);
		snooze(20.0 * 1000.0);
	}
	AboutRequested();
}
/*------------------------------------------------------------*/


void TMy3d::SetShape(TShape *a_shape)
{
	main_shape = a_shape;
}


/*------------------------------------------------------------*/

long TMy3d::_mouse_watcher(void *data)
{
	TMy3d *t = (TMy3d *)data;
	
	while( !t->dying ) {
		t->MouseWatcher();
		snooze(30.0 * 1000.0);
	}
	return 0;
}

/*------------------------------------------------------------*/

const float R_SQ = 400 * 400;

void	TMy3d::MouseWatcher()
{
	BPoint where;
	ulong buttons;
	bool putLock = false;
	
	if( Window( ) != NULL ) {
		if(_getmouse) {
			if(!Window()->IsLocked()) {
				Window( )->Lock( );
				putLock = true;
			}
			GetMouse(&where, &buttons, false);
			ConvertToScreen(&where);
			mouse_point = where;
			ConvertFromScreen(&where);
			if(putLock) {
				Window( )->Unlock( );
				putLock = false;
			}
		}
		else {
			where = mouse_point;
			if(!Window()->IsLocked()) {
				Window()->Lock();
				putLock = true;
			}
			ConvertFromScreen(&where);
			if(putLock) {
				Window()->Unlock();
				putLock = false;
			}
		}
		
		if(where == old_where) {
			idle_count++;
			if( idle_count > 500) {	//start the bored rotation
				if(bored_axis & 1)
					alpha += 0.05;
				if(bored_axis & 2)
					beta += 0.05;
				if(bored_axis & 4)
					gamma += 0.05;
				clear();
				main_shape->draw(alpha,
								 beta,
								 gamma,
								( rect.left + rect.right ) / 2, ( rect.top + rect.bottom ) / 2);
				if(!Window()->IsLocked()) {
					Window()->Lock();
					putLock = true;
				}
				Draw();		
				if(putLock) {
					Window()->Unlock();
					putLock = false;
				}
			} else {
				bored_axis = rand() % 4 + 1;
			}		
		} else {
			old_where = where;
			idle_count = 0;
	
			double x = where.x - ( rect.left + rect.right ) / 2;
			double y = where.y - ( rect.top + rect.bottom ) / 2;
			if( y * y > R_SQ )
				beta = M_PI_2;
			else	{
				double denominator = sqrt(R_SQ * ( R_SQ - y * y ));	
				double numerator = R_SQ - y * y ;
				beta = acos( ( R_SQ - (y * y) ) / sqrt( R_SQ * ( R_SQ - y * y ) ) );
			}
			
			if( beta > M_PI_2 * 0.75 )
				beta = M_PI_2 * 0.75;
			if( y > 0 )
				beta *= -1;
			if( x * x > R_SQ )
				gamma = M_PI_2;
			else	{
				float numerator = R_SQ - x * x;
				float denominator = sqrt( R_SQ * ( R_SQ - (x * x) )) ;
				gamma = acos(numerator/denominator);
			}
			if( gamma > M_PI_2 * 0.75 )
				gamma = M_PI_2 * 0.75;
			if( x > 0 )
				gamma *= -1;
			clear();
			main_shape->draw(alpha,
							 beta,
							 gamma,
							( rect.left + rect.right ) / 2, ( rect.top + rect.bottom ) / 2);
			if(!Window()->IsLocked()) {
				Window()->Lock();
				putLock = true;
			}
			Draw();
			if(putLock) {
				Window()->Unlock();
				putLock = false;
			}
				
		}
	}
}

void TMy3d::MessageReceived(BMessage *an_event)
{

	switch(an_event->what) {
		case B_QUIT_REQUESTED:
			be_app->PostMessage(new BMessage(B_QUIT_REQUESTED));
			break;
		case B_ABOUT_REQUESTED:
			AboutRequested();
			break;
		default:
			BView::MessageReceived(an_event);	
			break;
	}
}

/*------------------------------------------------------------*/
void TMy3d::AboutRequested()
{
	BAlert *alert = new BAlert("",
			"BeEyes\nCopyright 1998 ABiSoft, Inc. (bedev@abisoft.com)\nBased on source by Benoit Shillings\nwhich bears a copyright 1993, Be, Inc.\n",
			"Okay");
	alert->Go();
}

void TMy3d::MainView()
{
	_getmouse = true;
}

void add_eye(TShape *a_shape,
				 float center_x,
				 float center_y,
				 float center_z,
				 float size)
{
	long first;
	float x, y, z, radius;
	
	first =
	a_shape->add_point(center_x, center_y, center_z+size);
	
	for( int j = 0; j < SPH_GRAIN; j++ ) {
		for( int i = 0; i < CIR_GRAIN; i++ ) {
			radius = size * (SPH_GRAIN - j) / SPH_GRAIN;
			x = radius * sin(2 * M_PI * i / CIR_GRAIN);
			y = radius * cos(2 * M_PI * i / CIR_GRAIN);
			z = sqrt( size*size - radius * radius );

			a_shape->add_point(center_x + x, center_y + y, center_z + z);
		}
	}
	
	first++;
	for( int i = 0; i < SPH_GRAIN - 1; i++ ) {
		rgb_color col;
		if( i < SPH_GRAIN - 3 )
			col = make_color(220, 220, 220);
		else {
			col = make_color(65, 40, 255);
		}
		for( int j = 0; j < CIR_GRAIN - 1; j++ ) {
			a_shape->add_poly(	first + i * CIR_GRAIN + j,
								first + i * CIR_GRAIN + j + 1,
								first + (i + 1) * CIR_GRAIN + j + 1,
								first + (i + 1) * CIR_GRAIN + j,
								col);
		}
		a_shape->add_poly(	first + i * CIR_GRAIN + CIR_GRAIN - 1,
							first + i * CIR_GRAIN,
							first + (i + 1) * CIR_GRAIN,
							first + (i + 1) * CIR_GRAIN + CIR_GRAIN - 1,
							col);
	}

	for( int j = 0; j < CIR_GRAIN - 1; j++ ) {
		a_shape->add_poly(	first + (SPH_GRAIN - 1) * CIR_GRAIN + j,
							first + (SPH_GRAIN - 1) * CIR_GRAIN + j + 1,
							first - 1,
							first - 1,
							make_color(32, 248, 32));
	}
}

rgb_color make_color(uint8 r, uint8 g, uint8 b)
{
	rgb_color	c;
	
	c.red = r;
	c.green = g;
	c.blue = b;
	c.alpha = 255;
	return(c);
}
